/*
**    TP ETI5 CPE Lyon
**    Copyright (C) 2010 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <texture.hpp>

#include <png.h>
#include <errno.h>

#include <fstream>
#include <string_helper.hpp>

namespace cpe
{
    texture::texture()
        :N_x(0),N_y(0)
    {}

    void texture::load_picture(const std::string& filename)
    {
        std::string extension=filename.substr(filename.rfind('.'),filename.size());

        if(extension==".ppm")
            load_ppm_picture(filename);
        else if(extension==".png")
            load_png_picture(filename);
        else
        {
            std::cout<<"texture::load_picture unkown format: "<<extension<<std::endl;
            exit(-1);
        }
    }



    void texture::load_ppm_picture(const std::string& filename)
    {
        picture.clear();

        std::ifstream stream(filename.c_str());
        if(!stream.good())
            throw std::string("texture::load_ppm_picture("+filename+") cannot be opened");

        //go to magic number
        std::string buffer;
        buffer="#";
        while(buffer[0]=='#')
            std::getline(stream,buffer);
        if(buffer!="P3")
            throw std::string("texture::load_ppm_picture("+filename+") did not find P3 magic number: "+buffer);

        //Nx
        buffer="#";
        while(buffer[0]=='#')
            std::getline(stream,buffer);
        std::vector<std::string> size=cpe::string_converter::delete_empty(cpe::string_tokenizer::tokenize(buffer));
        if(size.size()<2)
            throw std::string("texture::load_ppm_picture("+filename+") cannot read size: "+buffer);
        N_x=cpe::string_converter::value_of<unsigned int>(size[0]);
        N_y=cpe::string_converter::value_of<unsigned int>(size[1]);

        //color number
        stream>>buffer;



        for(unsigned int k_x=0;k_x<N_x;k_x++)
        {
            for(unsigned int k_y=0;k_y<N_y;k_y++)
            {
                for(unsigned int k_dim=0;k_dim<3;k_dim++)
                {
                    stream>>buffer;
                    unsigned int val=cpe::string_converter::value_of<unsigned int>(buffer);
                    picture.push_back(val);
                }
                picture.push_back(255);
            }
        }


        //only store RGBA values
        depth=4;

        stream.close();

    }

    void texture::export_ppm(const std::string& filename) const
    {
        std::ofstream stream(filename.c_str(),std::ofstream::out);
        if(!stream.good())
            throw std::string("Error in Texture::export_ppm("+filename+"...) : cannot open file \n");

        //magic number
        stream<<"P3 \n";
        //size
        stream<<N_x<<" "<<N_y<<" \n";
        //color number
        stream<<"255"<<std::endl;

        for(unsigned int k2=0;k2<N_x;++k2)
        {
            for(unsigned int k1=0;k1<N_y;++k1)
            {
                stream<<static_cast<unsigned int>(picture[4*(N_y*k2+k1)+0])<<" "<<
                        static_cast<unsigned int>(picture[4*(N_y*k2+k1)+1])<<" "<<
                        static_cast<unsigned int>(picture[4*(N_y*k2+k1)+2])<<std::endl;
            }
        }
        stream.close();
    }

    const unsigned int& texture::get_Nx() const
    {
        return N_x;
    }
    const unsigned int& texture::get_Ny() const
    {
        return N_y;
    }
    const std::vector<unsigned char>& texture::get_data() const
    {
        return picture;
    }
    const unsigned int& texture::get_depth() const
    {
        return depth;
    }



    void texture::load_png_picture(const std::string& filename)
    {
        int             num_passes;   /* Number of interlace passes in file */
        int             pass;         /* Current pass in file */
        FILE           *fp;           /* File pointer */
        png_structp     pp;           /* PNG read pointer */
        png_infop       info;         /* PNG info pointers */
        unsigned char **pixels;       /* Pixel rows */

        pp = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
        info = png_create_info_struct (pp);

        if (setjmp (pp->jmpbuf))
          {
            printf ("Error while reading '%s'. File corrupted?\n", filename.c_str());
            return;
          }

        /* Open the file and initialize the PNG read "engine"...  */
        fp = fopen (filename.c_str (), "r");
        if (!fp)
          {
            printf ("Could not open '%s' for reading: %s\n", filename.c_str(), strerror (errno));
            return;
          }
        png_init_io (pp, fp);
        png_read_info (pp, info);

        if (info->bit_depth == 16)
          png_set_strip_16 (pp);

        if (info->color_type == PNG_COLOR_TYPE_GRAY && info->bit_depth < 8)
          png_set_expand (pp);

        /* Expand G+tRNS to GA, RGB+tRNS to RGBA */

        if (info->valid & PNG_INFO_tRNS)
          png_set_expand (pp);

        /*
         * Turn on interlace handling... libpng returns just 1 (ie single pass)
         * if the image is not interlaced
         */

        num_passes = png_set_interlace_handling (pp);

        /* Update the info structures after the transformations take effect */
        png_read_update_info (pp, info);

        N_x  = info->width;
        N_y = info->height;

        switch (info->color_type)
          {
          case PNG_COLOR_TYPE_RGB:           /* RGB */
            depth = 3;
            //type = IMAGE_TYPE_RGB;
            break;

          case PNG_COLOR_TYPE_RGB_ALPHA:     /* RGBA */
            depth = 4;
            //type = IMAGE_TYPE_RGBA;
            break;

          case PNG_COLOR_TYPE_GRAY:          /* Grayscale */
            depth = 1;
            //type = IMAGE_TYPE_GRAY;
            break;

          case PNG_COLOR_TYPE_GRAY_ALPHA:    /* Grayscale + alpha */
            depth = 2;
            //type = IMAGE_TYPE_GRAYA;
            break;

          default:                   /* Aie! Unknown type */
            printf ("Unknown color model in PNG file '%s'.\n", filename.c_str());
            return;
          }
        unsigned char *data = new unsigned char[N_x * N_y * depth];

        /* png decodes by row... */
        pixels = new unsigned char *[info->height];

        for (unsigned int i = 0; i < N_y; i++)
          pixels[i] = data + info->width * info->channels * i;

        for (pass = 0; pass < num_passes; pass++)
          png_read_rows (pp, pixels, 0, N_y);

        png_read_end (pp, info);

        png_destroy_read_struct (&pp, &info, 0);

        delete[] pixels;
        free (pp);
        free (info);

        fclose (fp);

        picture.resize(3*N_x*N_y);
        for(unsigned int k1=0;k1<N_x;++k1)
        {
            for(unsigned int k2=0;k2<N_y;++k2)
            {
                picture[3*(k1+N_x*k2)+0]=data[3*(k1+N_x*k2)+0];
                picture[3*(k1+N_x*k2)+1]=data[3*(k1+N_x*k2)+1];
                picture[3*(k1+N_x*k2)+2]=data[3*(k1+N_x*k2)+2];
            }
        }

    }
}
